import pytest
from pretty_tables import PrettyTables, TableColors


def test_generate_table(headers, rows):
    pretty_table = PrettyTables.generate_table(headers, rows, 'No data')
    assert pretty_table == ("| ID | Name   | Occupation        | Employed |\n"
                            "| 1  | Justin | Software Engineer | True     |\n"
                            "| 2  | Misty  | Receptionist      | False    |\n"
                            "| 3  | John   | No data           | False    |")


def test_generate_table_with_colors(headers, rows, colors):
    pretty_table = PrettyTables.generate_table(headers, rows, 'No data', colors)
    assert pretty_table ==(
           '| \x1b[94mID\x1b[0m | \x1b[95mName  \x1b[0m | \x1b[1mOccupation       \x1b[0m '
           '| \x1b[92mEmployed\x1b[0m |\n'
           '| \x1b[94m1 \x1b[0m | \x1b[95mJustin\x1b[0m | \x1b[1mSoftware Engineer\x1b[0m '
           '| \x1b[92mTrue    \x1b[0m |\n'
           '| \x1b[94m2 \x1b[0m | \x1b[95mMisty \x1b[0m | \x1b[1mReceptionist     \x1b[0m '
           '| \x1b[92mFalse   \x1b[0m |\n'
           '| \x1b[94m3 \x1b[0m | \x1b[95mJohn  \x1b[0m | \x1b[1mNo data          \x1b[0m '
           '| \x1b[92mFalse   \x1b[0m |'
    )

def test_generate_table_with_default_truthy(headers, rows):
    pretty_table = PrettyTables.generate_table(headers, rows, 'No data', truthy=3)
    assert pretty_table == (
           '| ID\x1b[0m | Name  \x1b[0m | Occupation       \x1b[0m | Employed\x1b[0m |\n| \x1b[92m1 \x1b[0m | '
           '\x1b[92mJustin\x1b[0m | \x1b[92mSoftware Engineer\x1b[0m | \x1b[92mTrue    '
           '\x1b[0m |\n| \x1b[91m2 '
           '\x1b[0m | \x1b[91mMisty \x1b[0m | \x1b[91mReceptionist     \x1b[0m | \x1b[91mFalse   '
           '\x1b[0m |\n| \x1b[91m3 \x1b[0m | '
           '\x1b[91mJohn  \x1b[0m | \x1b[91mNo data          \x1b[0m | \x1b[91mFalse   \x1b[0m |'
    )

def test_generate_table_with_default_truthy_custom(headers, rows):
    pretty_table = PrettyTables.generate_table(headers, rows, 'No data',
                                               colors=[TableColors.OKCYAN, TableColors.HEADER], truthy=3)
    assert pretty_table == (
           '| ID\x1b[0m | Name  \x1b[0m | Occupation       \x1b[0m | Employed\x1b[0m |\n'
           '| \x1b[96m1 \x1b[0m | \x1b[96mJustin\x1b[0m | \x1b[96mSoftware Engineer\x1b[0m | '
           '\x1b[96mTrue    \x1b[0m |\n'
           '| \x1b[95m2 \x1b[0m | \x1b[95mMisty \x1b[0m | \x1b[95mReceptionist     \x1b[0m | '
           '\x1b[95mFalse   \x1b[0m |\n'
           '| \x1b[95m3 \x1b[0m | \x1b[95mJohn  \x1b[0m | \x1b[95mNo data          \x1b[0m | '
           '\x1b[95mFalse   \x1b[0m |'
    )


def test_generate_table_with_default_truthy_not_enough_colors(headers, rows):
    with pytest.raises(ValueError) as exc:
        _ = PrettyTables.generate_table(headers, rows, 'No data',
                                        colors=[TableColors.OKCYAN], truthy=3)
    assert 'When using the truthy option you must either specify two colors, or no colors' in str(exc.value)


def test_generate_table_with_default_truthy_bad_column_index(headers, rows):
    with pytest.raises(ValueError) as exc:
        _ = PrettyTables.generate_table(headers, rows, 'No data', truthy='bad')
    assert 'The column specified for truthy values does not exist. Column bad' in str(exc.value)


def test_generate_table_with_default_truthy_out_of_range_column(headers, rows):
    with pytest.raises(ValueError) as exc:
        _ = PrettyTables.generate_table(headers, rows, 'No data', truthy=5)
    assert 'The column specified for truthy values does not exist. Column 5' in str(exc.value)


def test_validate_table_input_no_headers(rows):
    headers = None
    with pytest.raises(ValueError) as error:
        PrettyTables._validate_table_input(headers, rows)
    assert 'Headers are either not set or are not a proper array.' in str(error.value)


def test_validate_table_input_bad_headers(rows):
    headers = '123'
    with pytest.raises(ValueError) as error:
        PrettyTables._validate_table_input(headers, rows)
    assert 'Headers are either not set or are not a proper array.' in str(error.value)


def test_validate_table_input_no_rows(headers):
    rows = None
    with pytest.raises(ValueError) as error:
        PrettyTables._validate_table_input(headers, rows)
    assert 'Rows are either not set or are not a proper array.' in str(error.value)


def test_validate_table_input_bad_rows(headers):
    rows = '123'
    with pytest.raises(ValueError) as error:
        PrettyTables._validate_table_input(headers, rows)
    assert 'Rows are either not set or are not a proper array.' in str(error.value)


def test_validate_table_input_bad_colors(headers, rows):
    with pytest.raises(ValueError) as error:
        PrettyTables._validate_table_input(headers, rows, colors='asdf')
    assert 'Colors are set but are not a proper array.' in str(error.value)


def test_validate_table_input_colors_bad_length(headers, rows):
    with pytest.raises(IndexError) as error:
        PrettyTables._validate_table_input(headers, rows, colors=[1, 2])
    assert 'The number of colors does not match the number of columns.' in str(error.value)


def test_validate_table_input_bad_row_in_rows():
    headers = ['column1']
    rows = [
        '123',
    ]
    with pytest.raises(IndexError) as error:
        PrettyTables._validate_table_input(headers, rows)
    assert 'Row 1 is not a proper array.' in str(error.value)


def test_validate_table_input_mismatching_column_length():
    headers = ['column1', 'column2']
    rows = [
        ['123']
    ]
    with pytest.raises(IndexError) as error:
        PrettyTables._validate_table_input(headers, rows)
    assert 'Row 1 has 1 column(s) which does not match the table columns of 2.' in str(error.value)



"""The tables module for the pretty_tables library.
"""


class TableColors:
    """A class to contain constants for colors for the table.
    """

    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


class PrettyTables(object):
    """A class used to generate pretty tables for console output.
    """

    @classmethod
    def generate_table(cls, headers, rows, empty_cell_placeholder=None, colors=None, truthy=None):
        """Generate pretty tables with headers and rows (arrays).

        We do this by validating the headers and rows are valid arrays
        and are set, then convert all items to strings and append
        them to their respective row, then append each row to the table.
        We replace empty items with a placeholder and finally return
        the validated, generated, formatted table.

        Args:
            headers (list): a list of headers for the table.
            rows (list[list]): A list of rows for the table.
            empty_cell_placeholder: The value to be placed in cells with no data.
            colors (list): A list of TableColors for each row.
            truthy (int): Index of the column to be compared for truthy values. If truthy is not None
                truth comparison will override all other formats.
        """
        PrettyTables._validate_table_input(headers, rows)
        table = [headers]
        for row in rows:
            formatted_row = []
            for item in row:
                if item is None:
                    item = empty_cell_placeholder
                formatted_row.append(item)
            table.append(formatted_row)
        formatted_table = cls._format_table(table, colors, truthy)
        return formatted_table

    @staticmethod
    def _format_table(table, colors=None, truthy=None):
        """Take table data and format it into a pretty table.

        This includes adding lines for columns and uniform spacing.

        Args:
            table: The list defining the table to be formatted.
            colors (list): A list of TableColors for each row. If truthy is set to True, colors should
                be set to None, or a list of colors length 2, where the first color is to be used for
                true values and the second color is to be used for false values.
            truthy (int): Index of the column to be compared for truthy values. If truthy is not None
                truth comparison will override all other formats.
        """
        complete_table = []
        col_widths = [max(len(str(item)) for item in column) for column in zip(*table)]
        for line_no, line in enumerate(table):
            # Check for truthyness and colors.
            if colors or truthy:
                # Check if truthy
                if truthy:
                    # Check has the right number of colors for truthy coloring.
                    if colors and (len(colors) != 0 and len(colors) != 2):
                        raise ValueError(
                            'When using the truthy option you must either specify two colors, or no colors'
                        )
                    # Check the truthy column exist.
                    if not isinstance(truthy, int) or not 0 < truthy < len(line):
                        raise ValueError(
                            f'The column specified for truthy values does not exist. Column {truthy}'
                        )
                    # Default truthy colors if no colors are specified.
                    if not colors or len(colors) == 0:
                        colors = [TableColors.OKGREEN, TableColors.FAIL]
                    complete_table.append(
                        '| ' + ' | '.join(
                            f'{"" if line_no == 0 else colors[0] if bool(line[truthy]) else colors[1]}'
                            f'{str(item):{col_widths[i]}}{TableColors.ENDC}'
                            for i, item in enumerate(line)
                        ) + ' |'
                    )
                else:
                    # Set colors.
                    complete_table.append('| ' + ' | '.join(f'{colors[i]}{str(item):{col_widths[i]}}{TableColors.ENDC}'
                                                            for i, item in enumerate(line)) + ' |')
            else:
                complete_table.append('| ' + ' | '.join('{:{}}'.format(str(item), col_widths[i])
                                                        for i, item in enumerate(line)) + ' |')

        formatted_table = '\n'.join(complete_table)
        return formatted_table

    @staticmethod
    def _validate_table_input(headers, rows, colors=None):
        """Validate table input.

        Checking that headers and rows are set and each are a
        valid list and correct length.

        Args:
            headers (list): A list of headers for the table.
            rows (list): A list of lists matching the length of headers for each row.
            colors (list): A list of TableColors for each column.
        """
        if not headers or not isinstance(headers, list):
            raise ValueError(
                'Headers are either not set or are not a proper array.'
            )
        elif not rows or not isinstance(rows, list):
            raise ValueError(
                'Rows are either not set or are not a proper array.'
            )
        if colors:
            if not isinstance(colors, list):
                raise ValueError(
                    'Colors are set but are not a proper array.'
                )
            elif len(colors) != len(headers):
                raise IndexError('The number of colors does not match the number of columns.')
        table_length = len(headers)
        for i, row in enumerate(rows):
            if not isinstance(row, list):
                raise IndexError(
                    f'Row {i + 1} is not a proper array.'
                )
            row_length = len(row)
            if row_length != table_length:
                raise IndexError(
                    f'Row {i + 1} has {row_length} column(s) which does not match the table columns of {table_length}.'
                )

        return True



"""Fixtures for PrettyTable testing"""
import pytest

from pretty_tables import TableColors


@pytest.fixture(scope='function')
def headers():
    return ['ID', 'Name', 'Occupation', 'Employed']


@pytest.fixture(scope='function')
def rows():
    return [
        [1, 'Justin', 'Software Engineer', True],
        [2, 'Misty', 'Receptionist', False],
        [3, 'John', None, False],
    ]


@pytest.fixture(scope='function')
def colors():
    return [TableColors.OKBLUE, TableColors.HEADER, TableColors.BOLD, TableColors.OKGREEN]
